from helper_code import *
import numpy as np
import scipy.signal, scipy.io, scipy
import csv
import math
#import librosa

def get_raw(data_folder, train=False, verbose=0):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    outcome_classes = ['Abnormal', 'Normal']
    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    features = list()
    murmurs = list()
    outcomes = list()
    recordings = list()

    for i in range(num_patient_files):
        if verbose >= 2:
            print('    {}/{}...'.format(i+1, num_patient_files))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)
        recordings.append(current_recordings)

        # Extract features.
        if not train:
            current_features = get_test_features(current_patient_data)
        else:
            current_features = get_train_features(current_patient_data, data_folder)
        features.append(current_features)

        # Extract labels and use one-hot encoding.
        murmur = murmur_classes.index(get_murmur(current_patient_data))
        murmurs.append(murmur)
        outcome = outcome_classes.index(get_outcome(current_patient_data))
        outcomes.append(outcome)
    return recordings, features, murmurs, outcomes

def get_test_features(data):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = 0.0

        # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_feature = 0
    if compare_strings(sex, 'Female'):
        sex_feature = 1
    elif compare_strings(sex, 'Male'):
        sex_feature = 2

    l = data.split('\n')[0]
    frequency = str(l.split('\t')[2].strip())

    # Extract height and weight.
    height = get_height(data)
    if not height or np.isnan(height):
        height = 0.0
    weight = get_weight(data)
    if not weight or np.isnan(weight):
        weight = 0.0

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    locations = get_locations(data)

    features = [age, sex_feature, height, weight, is_pregnant, locations, frequency]
    return features

def get_train_features(data, data_folder):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = 0.0

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_feature = 0
    if compare_strings(sex, 'Female'):
        sex_feature = 1
    elif compare_strings(sex, 'Male'):
        sex_feature = 2

    l = data.split('\n')[0]
    frequency = str(l.split(' ')[2].strip())

    # Extract height and weight.
    height = get_height(data)
    if not height or np.isnan(height):
        height = 0.0
    weight = get_weight(data)
    if not weight or np.isnan(weight):
        weight = 0.0

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    locations = get_locations(data)

    # The following data is only available for training data
    mal = None
    for l in data.split('\n'):
        if l.startswith('#Most audible location:'):
            mal = str(l.split(': ')[1].strip())

    mlocs = []
    for l in data.split('\n'):
        if l.startswith('#Murmur locations:'):
            loc_str = str(l.split(': ')[1].strip())
            mlocs = loc_str.split('+')

    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations + 1]
    segmentations = list()
    for i in range(num_locations):
        segmentation = list()
        entries = recording_information[i].split(' ')
        try:
            recording_seg = entries[3]
        except:
            segmentations.append([None])
            continue
        filename = os.path.join(data_folder, recording_seg)
        with open(filename) as tsv:
            for line in csv.reader(tsv, dialect="excel-tab"):
                segmentation.append((line[0], line[2]))
        segmentations.append(segmentation)

    features = [age, sex_feature, height, weight, is_pregnant, locations,frequency, segmentations, mal, mlocs]
    return features

def extract_harmonic(record):
    # removes the percussive part of the recording
    return librosa.effects.harmonic(record)

def preprocess_individual_recording(record, freq, length=5000, harmonic = False, cycle = False):
    band = butterworth(record, order=6, low=25, high=400, freq=freq)
    if harmonic:
        band = extract_harmonic(band)
    if cycle:
        interv = get_heart_cycle_intervals(recording=band, sampling_rate=freq, threshold = 0.01, min_width = 2000)
        band = band[interv[0][0]:interv[2][0]]
    down = downsample(band, 4)
    down_norm = minmax_normalise(down)
    padded_out = pad(down_norm, length)
    # sampling freq now 1 kHz
    return padded_out

def mark_first_cycle(features):
    # the last cardiac cycle happens between 1.4 and 10 seconds
    # the first happens between 0 and 10 seconds
    if not isinstance(features[0][-3][0][0], tuple):
        raise ValueError("No segmentation in features")

    starts=list()
    ends=list()
    for p in features:
        for seg in p[-3]:
            first = False
            seg.reverse()
            for t in seg:
                if not t:
                    continue
                if int(t[1]) == 1 and not first:
                    starts.append(t[0])
                    first = True
                elif int(t[1]) == 1:
                    ends.append(t[0])
                    break
    print(len(starts))
    print(len(ends))
    print(max(starts))
    print(min(ends))

def butterworth(recording, order, low, high, freq):
    b1, a1 = scipy.signal.butter(order, (low, high),"bandpass", fs=float(freq), output='ba')
    bandpassed = scipy.signal.lfilter(b1, a1, recording)
    return np.array(bandpassed).astype(np.float32)

def mean_downsample(recording, factor):
    pad_size = math.ceil(float(recording.size) / factor) * factor - recording.size
    b_padded = np.append(recording, np.zeros(pad_size) * np.NaN)
    res = np.nanmean(b_padded.reshape(-1, factor), axis=1)
    return np.array(res).astype(np.float32)

def downsample(recording, factor):
    return scipy.signal.decimate(recording, factor)

def minmax_normalise(recording):
    min_v = int(recording.min())
    max_v = int(recording.max())
    res = (recording - min_v)/ (max_v - min_v)
    return np.array(res).astype(np.float32)

def pad(audio, length):
    # Zero-padding for an audio waveform with less than input_len samples.
    waveform = audio[:length]
    zero_padding = np.zeros(length - len(waveform), dtype=int)
    # Concatenate the waveform with `zero_padding`, which ensures all audio
    # clips are of the same length.
    equal_length = np.concatenate([waveform, zero_padding], 0)
    return equal_length

### stolen (adapted) from Tim :D

def envelope_peaks(
        envelope: np.ndarray,
        threshold: float = 0.05,
        min_width=2,
        ):
    envelope_norm = envelope - envelope.min()
    envelope_norm /= envelope_norm.max() + 1e-8
    peaks, props = scipy.signal.find_peaks(
        envelope_norm, prominence=threshold, width=2, distance=min_width
    )
    onset = props["left_ips"]
    peaks = peaks.astype(np.int32)
    onset = onset.astype(np.int32)

    peaks.sort()
    return peaks, onset

def get_heart_cycle_intervals(
        recording: np.ndarray,
        sampling_rate: int,
        threshold: float = 0.01,
        min_width: int = 2000,
        ):
    rec_le = butterworth(recording, order=2, low=0.01, high=4, freq=sampling_rate)
    peaks, onsets = envelope_peaks(np.abs(rec_le), threshold=threshold, min_width=min_width)
    offsets = np.array(onsets[1:].tolist() + [len(peaks)])

    inters = list(zip(onsets, offsets))
    return list(filter(lambda x: x[1] - x[0] > 0, inters))

if __name__ == "__main__":
    recordings, features, murmurs, outcomes = get_raw(data_folder="/home/sven/Desktop/physionet2/code/ukj_fsu_physionet22/data/training_data/cv_test",
                                                      train=True)
    for i,p in enumerate(recordings):
        if murmurs[i] == 0:
            cycle = preprocess_individual_recording(p[0], freq=4000)
            break
    scipy.io.wavfile.write("/home/sven/Desktop/physionet2/code/ukj_fsu_physionet22/sample.wav", rate=1000, data=cycle)