import tensorflow as tf
import matplotlib.pyplot as plt


class Categor_Focal_Loss(tf.keras.losses.Loss):

    def __init__(self, alpha, gamma, reduction):
        super().__init__(reduction=reduction)
        self.alpha = alpha
        self.gamma = gamma

    def call(self, y_true, y_pred):
        #loss = -alpha*((1-p)^gamma)*log(p)
        y_pred = tf.nn.softmax(y_pred)
        y_true = tf.cast(tf.one_hot(y_true, depth=3), dtype=tf.dtypes.float32)
        epsilon = 1e-10
        y_pred = tf.clip_by_value(y_pred, epsilon, 1.0 - epsilon)
        ce = -y_true * tf.math.log(y_pred)
        weight = self.alpha * y_true * tf.math.pow((1 - y_pred), self.gamma)
        loss = tf.reduce_sum(weight * ce, axis=-1)
        return loss


class Model(tf.keras.Model):

    def create_atrous_conv_layer(self, num_filters, size, dilation):
        conv = tf.keras.layers.Conv1D(num_filters, size, strides=1, padding='valid',
                                      data_format='channels_last', dilation_rate=dilation, groups=1,
                                      activation="elu", kernel_initializer='glorot_uniform')
        return conv
    def create_lstm_layer(self, hidden_state_dimension,sequence=True, direction="unidirectional", name=""):
        if direction == "bidirectional":
            hidden_state_dimension = int(hidden_state_dimension / 2)
        lstm_layer = tf.keras.layers.LSTM(units=hidden_state_dimension,
                                        recurrent_activation='sigmoid', return_sequences=sequence, return_state=False,name="forward_lstm_{}".format(name))

        if direction == "bidirectional":
            # already returns the reversed sequence
            bwd_lstm_layer = tf.keras.layers.LSTM(units=hidden_state_dimension,
                                                recurrent_activation='sigmoid', go_backwards=True,
                                                return_sequences=sequence, return_state=False, name="backward_lstm_{}".format(name))
            lstm_layer = tf.keras.layers.Bidirectional(layer=lstm_layer, merge_mode='concat',
                                                      backward_layer=bwd_lstm_layer, name="full_bi_lstm_{}".format(name))

        return lstm_layer

    def produce_wav_to_spectro(self, fov, stride):
        def wav_to_spectrogram(binary):
            # Convert the waveform to a spectrogram via a STFT.
            # window and step size need to be computed from the convolution parameters (field of view // stride)
            window = tf.constant(fov) #298
            step = tf.constant(stride) #31
            # time bins = ceil((time_steps-window)/stride)+1
            spectrogram = tf.stop_gradient(tf.signal.stft(
                binary, frame_length=window, frame_step=step, fft_length=200, pad_end=False))
            # Obtain the magnitude of the STFT.
            spectrogram = tf.stop_gradient(tf.abs(spectrogram))
            # height [-2] corresponds to time bins and width [-1] corresponds to frequency bins
            # [batch size,  time bins (313), frequency bins]
            tr_spec = tf.transpose(spectrogram, [0, 2, 1])
            tr_spec_image = tf.expand_dims(tr_spec, 3)
            return tr_spec_image
        return wav_to_spectrogram

    @tf.function
    def call(self, recording, features, training=False):
        #x: [batch, time, 1]
        #features: [batch, 4]
        with tf.name_scope("murmur_head") as scope1:
            # pure record part
            x = tf.cast(tf.expand_dims(recording, axis=-1), tf.float32)
            # atrous temporal CNN
            for i, l in enumerate(self.atrconv_list):
                x = l(x)
                x = self.pooling_list[i](x)
                x = self.bn_list[i](x, training=training)
            # field of view: on average 26 on input (with max 2 neighbours ignored)
            # temporal CNN
            for j, l in enumerate(self.conv_list):
                x = l(x)
                x = self.pooling_list[2 + j](x)
                x = self.bn_list[2 + j](x, training=training)
            raw_flat = self.flatten(x)
            #print(raw_flat)
            # specto part
            specto = self.wav_to_spectogram(recording)
            #log_spec = tf.math.log(tf.transpose(spectro, [0,2,1])+1e-14)
            #plt.imshow(log_spec[0,:,:])
            #plt.show()
            x = specto
            for i, l in enumerate(self.conv_list2D):
                x = l(x)
                x = self.pooling_list2D[i](x)
                x = self.bn_list2D[i](x, training=training)
            specto_flat = self.flatten2(x)
            #print(specto_flat)
            raw_plus_spec = tf.concat([raw_flat, specto_flat], axis=-1)
            x = raw_plus_spec
            x = self.drop(x, training=training)
            x = self.dense2(self.dense(x))
            # add age,sex, height, weight
            x_plus_features = tf.concat([x,features], axis=-1)
            last_dense = self.dense3(x_plus_features)
            dec_murmur = self.decision_murmur(last_dense)
        with tf.name_scope("outcome_head") as scope2:
            den_out = self.dense_outcome_head(x_plus_features)
            dec_out = self.decision_outcome_head(den_out)
        return dec_murmur, dec_out

    def __init__(self):
        super(Model, self).__init__()
        self.atrconv_list = [self.create_atrous_conv_layer(64, 5, 2), self.create_atrous_conv_layer(32, 5, 2)]
        # field of view after the two atrous CNN: 26 (min:25, max:27), stride: 2 (min:1, max:3)
        self.conv_list = [self.create_atrous_conv_layer(2, 5, 1) for i in range(1)]
        # approx. field of view after the i-th (i=0: last atrous) additional CNN AND pooling (fov_0=25):
        # fov_{i-1}+overstride_{i-1}*(window_size_i)
        # approx. overstride (on original input) after the i-th (i=0: last atrous) additional CNN AND pooling (overstride_0=4):
        # poolwindow_i * overstride_{i-1}
        self.wav_to_spectogram = self.produce_wav_to_spectro(fov=40, stride=16)
        self.pooling_list = [tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding='valid') for i in range(2)] + [tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding='valid')]
        self.conv_list2D = [tf.keras.layers.Conv2D(16, (20,2), activation='elu'),
                            tf.keras.layers.Conv2D(8, (5,5), activation='elu')]
        self.pooling_list2D = [tf.keras.layers.MaxPooling2D(pool_size=4, strides=4, padding='valid') for i in range(2)]
        self.bn_list2D = [tf.keras.layers.BatchNormalization() for i in range(2)]
        self.bn_list = [tf.keras.layers.BatchNormalization() for i in range(5)]
        self.flatten = tf.keras.layers.Flatten()
        self.flatten2 = tf.keras.layers.Flatten()
        ## specto part
        self.drop = tf.keras.layers.Dropout(0.3)
        self.dense = tf.keras.layers.Dense(1024, activation="elu")
        self.dense2 = tf.keras.layers.Dense(512, activation="elu")
        self.dense3 = tf.keras.layers.Dense(256, activation="elu")
        self.decision_murmur = tf.keras.layers.Dense(3)
        #===== 2 head to predict outcome ====
        self.dense_outcome_head = tf.keras.layers.Dense(30, activation="elu")
        self.decision_outcome_head = tf.keras.layers.Dense(1)