import tensorflow as tf

class ResIDBlock(tf.keras.Model):
    def __init__(self, f1, f2, s):
        super(ResIDBlock, self).__init__()
        self.conv1 = tf.keras.layers.Conv1D(filters=f1, kernel_size=1, strides=1, padding='valid',
                            use_bias=False, kernel_initializer='normal')
        self.conv2 = tf.keras.layers.Conv1D(filters=f1, kernel_size=3, strides=1, padding='same',
                            use_bias=False, kernel_initializer='normal')
        self.conv3 = tf.keras.layers.Conv1D(filters=f2, kernel_size=1, strides=1, padding='valid',
                            use_bias=False, kernel_initializer='normal')
        #batch norm over channels (one mean for all time steps and batch-samples)
        self.batchnorm1 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.batchnorm2 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.batchnorm3 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.relu1 = tf.keras.layers.ReLU()
        self.relu2 = tf.keras.layers.ReLU()
        self.relu3 = tf.keras.layers.ReLU()

    def call(self, recording, training=False):
        x = recording
        x = self.conv1(x)
        x = self.batchnorm1(x, training=training)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.batchnorm2(x, training=training)
        x = self.relu2(x)
        x = self.conv3(x)
        x = self.batchnorm3(x, training=training)
        #skip
        x_plus_record = tf.add(x, recording)
        out = self.relu3(x_plus_record)
        return out

class ResConvBlock(tf.keras.Model):
    def __init__(self, f1,f2, s):
        super(ResConvBlock, self).__init__()
        self.conv1 = tf.keras.layers.Conv1D(filters=f1, kernel_size=1, strides=s, padding='valid',
                            use_bias=False, kernel_initializer='normal')
        self.conv2 = tf.keras.layers.Conv1D(filters=f1, kernel_size=3, strides=1, padding='same',
                            use_bias=False, kernel_initializer='normal')
        self.conv3 = tf.keras.layers.Conv1D(filters=f2, kernel_size=1, strides=1, padding='valid',
                            use_bias=False, kernel_initializer='normal')
        self.conv4 = tf.keras.layers.Conv1D(filters=f2, kernel_size=1, strides=s, padding='valid',
                            use_bias=False, kernel_initializer='normal')
        #batch norm over channels (one mean for all time steps and batch-samples)
        self.batchnorm1 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.batchnorm2 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.batchnorm3 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.batchnorm4 = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999, axis=-1)
        self.relu1 = tf.keras.layers.ReLU()
        self.relu2 = tf.keras.layers.ReLU()
        self.relu3 = tf.keras.layers.ReLU()
        self.relu4 = tf.keras.layers.ReLU()

    def call(self, recording, training=False):
        x = recording
        x = self.conv1(x)
        x = self.batchnorm1(x, training=training)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.batchnorm2(x, training=training)
        x = self.relu2(x)
        x = self.conv3(x)
        x = self.batchnorm3(x, training=training)
        #skip
        x_skip = self.conv4(recording)
        x_skip = self.batchnorm4(x_skip, training=training)
        x_plus_record = tf.add(x, x_skip)
        out = self.relu4(x_plus_record)
        return out

class Model(tf.keras.Model):

    def create_atrous_conv_layer(self, num_filters, size, dilation):
        conv = tf.keras.layers.Conv1D(num_filters, size, strides=1, padding='valid',
                                      data_format='channels_last', dilation_rate=dilation, groups=1,
                                      activation="elu", kernel_initializer='glorot_uniform')
        return conv
    def create_lstm_layer(self, hidden_state_dimension,sequence=True, direction="unidirectional", name=""):
        if direction == "bidirectional":
            hidden_state_dimension = int(hidden_state_dimension / 2)
        lstm_layer = tf.keras.layers.LSTM(units=hidden_state_dimension,
                                        recurrent_activation='sigmoid', return_sequences=sequence, return_state=False,name="forward_lstm_{}".format(name))

        if direction == "bidirectional":
            # already returns the reversed sequence
            bwd_lstm_layer = tf.keras.layers.LSTM(units=hidden_state_dimension,
                                                recurrent_activation='sigmoid', go_backwards=True,
                                                return_sequences=sequence, return_state=False, name="backward_lstm_{}".format(name))
            lstm_layer = tf.keras.layers.Bidirectional(layer=lstm_layer, merge_mode='concat',
                                                      backward_layer=bwd_lstm_layer, name="full_bi_lstm_{}".format(name))

        return lstm_layer

    def produce_wav_to_spectro(self, fov, stride):
        def wav_to_spectrogram(binary):
            # Convert the waveform to a spectrogram via a STFT.
            # window and step size need to be computed from the convolution parameters (field of view // stride)
            window = tf.constant(fov) #298
            step = tf.constant(stride) #31
            # time bins = ceil((time_steps-window)/stride)+1
            spectrogram = tf.stop_gradient(tf.signal.stft(
                binary, frame_length=window, frame_step=step, pad_end=False))
            # Obtain the magnitude of the STFT.
            spectrogram = tf.stop_gradient(tf.abs(spectrogram))
            # height [-2] corresponds to time bins and width [-1] corresponds to frequency bins
            # [batch size,  time bins (313), frequency bins]
            return spectrogram
        return wav_to_spectrogram

    @tf.function(experimental_relax_shapes=True)
    def call(self, recording, features, training=False):
        with tf.name_scope("murmur_head") as scope1:
            # residual
            x = tf.cast(tf.expand_dims(recording, axis=-1), tf.float32)
            # stage1
            x = self.conv_pure1(x)
            x = self.batchnorm_pure(x, training=training)
            x = self.relu_pure(x)
            x = self.max_pooling(x)
            for i, l in enumerate(self.resblock_list):
                x = l(x)
            # concat resnet output with spectrogram
            # [batch, time bins, frequency bins]
            spectro = self.wav_to_spectro(recording)
            x_con_spectro = tf.concat([x, spectro], axis=-1)
            x = x_con_spectro
            # biLSTMS
            for l in self.bi_lstm:
                x = l(x)
            x = self.drop(x, training=training)
            features = tf.cast(features, dtype=tf.float32)
            x_con_features = tf.concat([x,features], axis=-1)
            x = self.dense(x_con_features)
            dec_murmur = self.dense2(x)
        with tf.name_scope("outcome_head") as scope2:
            den_out = self.dense_outcome_head(x)
            dec_out = self.decision_outcome_head(den_out)
        return dec_murmur, dec_out

    def __init__(self):
        super(Model, self).__init__()
        self.conv_pure1 = tf.keras.layers.Conv1D(64, kernel_size=7 , strides=2)
        self.batchnorm_pure = tf.keras.layers.BatchNormalization()
        self.relu_pure = tf.keras.layers.ReLU()
        self.max_pooling = tf.keras.layers.MaxPooling1D(3, strides=2)
        filters_stride = [(64, 256, 1), (64, 256, -1), (64, 256, -1),
                          (128, 512,2), (128, 512, -1), (128, 512, -1), (128, 512, -1),
                          (256, 1024, 2), (256, 1024,-1), (256, 1024,-1), (256, 1024,-1), (256, 1024,-1), (256, 1024,-1),
                          (512, 2048, 2), (512, 2048,-1), (512, 2048, -1)]
        self.resblock_list = [ResIDBlock(f1,f2, 0) if s < 0 else ResConvBlock(f1,f2,s) for f1,f2,s in filters_stride]
        # fov is same as the effective fov of resnet50
        self.wav_to_spectro = self.produce_wav_to_spectro(fov=483, stride=29)
        self.bi_lstm = [self.create_lstm_layer(20, direction="bidirectional", sequence=False) for i in range(1)]
        self.drop = tf.keras.layers.Dropout(0.2)
        self.dense = tf.keras.layers.Dense(20, activation="relu")
        self.dense2 = tf.keras.layers.Dense(3)
        #===== 2 head to predict outcome ====
        self.dense_outcome_head = tf.keras.layers.Dense(30, activation="relu")
        self.decision_outcome_head = tf.keras.layers.Dense(1)