import numpy as np, os, sys
sys.path.append("..")
from helper_code import *
from team_code import load_challenge_model, run_challenge_model
import tensorflow as tf

# Run model.
def run_model(model_folder, data_folder, verbose):
    # Load models.
    if verbose >= 1:
        print('Loading Challenge model...')

    model = load_challenge_model(model_folder, verbose) ### Teams: Implement this function!!!

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')


    # Run the team's model on the Challenge data.
    if verbose >= 1:
        print('Running model on Challenge data...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    outcome_classes = ['Abnormal', 'Normal']
    murmur_labels = load_murmurs(patient_files, murmur_classes)
    outcome_labels = load_outcomes(patient_files, outcome_classes)
    overall_probs = np.zeros((num_patient_files,5))
    overall_decisions = np.zeros((num_patient_files,5))
    # Iterate over the patient files.
    for i in range(num_patient_files):
        if verbose >= 2:
            print('    {}/{}...'.format(i+1, num_patient_files))

        patient_data = load_patient_data(patient_files[i])
        recordings = load_recordings(data_folder, patient_data)

        # Allow or disallow the model to fail on parts of the data; helpful for debugging.
        classes, labels, probabilities = run_challenge_model(model, patient_data, recordings, verbose) ### Teams: Implement this function!!!
        overall_decisions[i] = labels
        overall_probs[i] = probabilities

    murmur_dict = dict()
    outcome_dict = dict()
    tweak_ps = [[[x/100, 0, y/100], [x/100]] for x in range(60) for y in range(60)]
    for tweak in tweak_ps:
        overall_probsl = overall_probs.tolist()
        murmur_dec = list()
        outcome_dec = list()
        if tweak[0] is not None:
            for i, probs in enumerate(overall_probsl):
                ensemble_prob_murmur = probs[:3]
                if ensemble_prob_murmur[0] >= tweak[0][0]:
                    max_murmur_index = 0
                    res_murmur = [1 if max_murmur_index == i else 0 for i in range(3)]
                    murmur_dec.append(res_murmur)
                elif ensemble_prob_murmur[2] >= tweak[0][2]:
                    max_murmur_index = 2
                    res_murmur = [1 if max_murmur_index == i else 0 for i in range(3)]
                    murmur_dec.append(res_murmur)
                else:
                    max_murmur_index = np.argmax(ensemble_prob_murmur)
                    res_murmur = [1 if max_murmur_index == i else 0 for i in range(3)]
                    murmur_dec.append(res_murmur)
            murmur_classes = ['Present', 'Unknown', 'Absent']
            murmur_score_valid = compute_weighted_accuracy(murmur_labels, np.array(murmur_dec), murmur_classes)
            murmur_dict[tuple(tweak[0])] = murmur_score_valid

        if tweak[1] is not None:
            for i, probs in enumerate(overall_probsl):
                ensemble_prob_outcome = probs[3:]
                if ensemble_prob_outcome[0] >= tweak[1][0]:
                    max_outcome_index = 0
                    res_outcome = np.array([1 if max_outcome_index == i else 0 for i in range(2)])
                    outcome_dec.append(res_outcome)
                else:
                    max_outcome_index = np.argmax(ensemble_prob_outcome)
                    res_outcome = [1 if max_outcome_index == i else 0 for i in range(2)]
                    outcome_dec.append(res_outcome)
            # outcomes_named = tf.where(overall_outcomes > 0.5, 'Normal', 'Abnormal')
            # decisions_named = tf.where(overall_decision > 0.5, 'Normal', 'Abnormal')
            outcome_score_valid = compute_challenge_metric(outcome_labels, np.array(outcome_dec), ['Abnormal', 'Normal'],
                                                           ['Abnormal', 'Normal'])
            outcome_dict[tuple(tweak[1])] = outcome_score_valid

    print(max(murmur_dict, key=murmur_dict.get), max(murmur_dict.values()))
    print(min(outcome_dict, key=outcome_dict.get), min(outcome_dict.values()))

def get_murmur_decision(logits):
    probabilities = tf.nn.softmax(logits).numpy()
    max_prob_index = np.argmax(probabilities, axis=-1)
    return max_prob_index, probabilities

def get_outcome_decision(logit):
    logit_np = logit.numpy()
    if logit_np >= 0:
        decision = 1
    else:
        decision = 0
    probabilities = [1-tf.sigmoid(logit).numpy(), tf.sigmoid(logit).numpy()]
    return decision, probabilities

# Compute a binary confusion matrix, where the columns are the expert labels and the rows are the classifier labels.
def compute_confusion_matrix(labels, outputs):
    assert(np.shape(labels)[0] == np.shape(outputs)[0])
    assert(all(value in (0, 1, True, False) for value in np.unique(labels)))
    assert(all(value in (0, 1, True, False) for value in np.unique(outputs)))

    num_patients = np.shape(labels)[0]
    num_label_classes = np.shape(labels)[1]
    num_output_classes = np.shape(outputs)[1]

    A = np.zeros((num_output_classes, num_label_classes))
    for k in range(num_patients):
        for i in range(num_output_classes):
            for j in range(num_label_classes):
                if outputs[k, i] == 1 and labels[k, j] == 1:
                    A[i, j] += 1

    return A

def compute_weighted_accuracy(labels, outputs, classes):
    # Define constants.
    if classes == ['Present', 'Unknown', 'Absent']:
        weights = np.array([[5, 3, 1], [5, 3, 1], [5, 3, 1]])
    elif classes == ['Abnormal', 'Normal']:
        weights = np.array([[5, 1], [5, 1]])
    else:
        raise NotImplementedError('Weighted accuracy undefined for classes {}'.format(', '.join(classes)))

    # Compute confusion matrix.
    assert(np.shape(labels) == np.shape(outputs))
    A = compute_confusion_matrix(labels, outputs)

    # Multiply the confusion matrix by the weight matrix.
    assert(np.shape(A) == np.shape(weights))
    B = weights * A

    # Compute weighted_accuracy.
    if np.sum(B) > 0:
        weighted_accuracy = np.trace(B) / np.sum(B)
    else:
        weighted_accuracy = float('nan')

    return weighted_accuracy

# Define total cost for algorithmic prescreening of m patients.
def cost_algorithm(m):
    return 10*m

# Define total cost for expert screening of m patients out of a total of n total patients.
def cost_expert(m, n):
    return (25 + 397*(m/n) -1718*(m/n)**2 + 11296*(m/n)**4) * n

# Define total cost for treatment of m patients.
def cost_treatment(m):
    return 10000*m

# Define total cost for missed/late treatement of m patients.
def cost_error(m):
    return 50000*m

# Compute Challenge metric.
def compute_challenge_metric(labels, outputs, label_classes, output_classes):
    # Define positive and negative classes for referral and treatment.
    positive_classes = ['Present', 'Unknown', 'Abnormal']
    negative_classes = ['Absent', 'Normal']

    # Compute confusion matrix.
    A = compute_confusion_matrix(labels, outputs)

    # Identify positive and negative classes for referral.
    idx_label_positive = [i for i, x in enumerate(label_classes) if x in positive_classes]
    idx_label_negative = [i for i, x in enumerate(label_classes) if x in negative_classes]
    idx_output_positive = [i for i, x in enumerate(output_classes) if x in positive_classes]
    idx_output_negative = [i for i, x in enumerate(output_classes) if x in negative_classes]

    # Identify true positives, false positives, false negatives, and true negatives.
    tp = np.sum(A[np.ix_(idx_output_positive, idx_label_positive)])
    fp = np.sum(A[np.ix_(idx_output_positive, idx_label_negative)])
    fn = np.sum(A[np.ix_(idx_output_negative, idx_label_positive)])
    tn = np.sum(A[np.ix_(idx_output_negative, idx_label_negative)])
    total_patients = tp + fp + fn + tn

    # Compute total cost for all patients.
    total_cost = cost_algorithm(total_patients) \
        + cost_expert(tp + fp, total_patients) \
        + cost_treatment(tp) \
        + cost_error(fn)

    # Compute mean cost per patient.
    if total_patients > 0:
        mean_cost = total_cost / total_patients
    else:
        mean_cost = float('nan')

    return mean_cost

def numpy_one_hot(a, max_ind):
    a = a.astype(np.int32)
    a_one_hot = np.zeros((a.size, max_ind))
    a_one_hot[np.arange(a.size), a] = 1
    return a_one_hot


# Load murmurs from label files.
def load_murmurs(label_files, classes):
    num_patients = len(label_files)
    num_classes = len(classes)

    # Use one-hot encoding for the labels.
    labels = np.zeros((num_patients, num_classes))

    # Iterate over the patients.
    for i in range(num_patients):
        data = load_patient_data(label_files[i])
        label = get_murmur(data)
        for j, x in enumerate(classes):
            if compare_strings(label, x):
                labels[i, j] = 1

    return labels

# Load outcomes from label files.
def load_outcomes(label_files, classes):
    num_patients = len(label_files)
    num_classes = len(classes)

    # Use one-hot encoding for the labels.
    labels = np.zeros((num_patients, num_classes))

    # Iterate over the patients.
    for i in range(num_patients):
        data = load_patient_data(label_files[i])
        label = get_outcome(data)
        for j, x in enumerate(classes):
            if compare_strings(label, x):
                labels[i, j] = 1

    return labels

if __name__ == '__main__':
    # Define the model, data, and output folders.
    model_folder = sys.argv[2]
    data_folder = sys.argv[1]

    # Allow or disallow the model to fail on parts of the data; helpful for debugging.
    allow_failures = False

    # Change the level of verbosity; helpful for debugging.
    if len(sys.argv)==5 and is_integer(sys.argv[4]):
        verbose = int(sys.argv[4])
    else:
        verbose = 1

    run_model(model_folder, data_folder, verbose)